#include <ros/ros.h>
#include <sensor_msgs/image_encodings.h>
#include <image_transport/image_transport.h>
#include <geometry_msgs/PoseStamped.h>
#include <cv_bridge/cv_bridge.h>
#include <tf/tf.h>
#include <tf/transform_listener.h>
#include <tf/transform_datatypes.h>

#include <iostream>
#include <thread>
#include <mutex>
#include <chrono>

#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>

#include "gridmapping.hpp"

static const std::string OPENCV_WINDOW = "Image window";

void createIceFeatureList(std::vector<AMOS::IceFeature>& list) {
	auto constRad = [](float x, float y) { return AMOS::IceFeature(x, y, 1); };
	list.push_back(constRad(0.0, 0.0));

	list.push_back(constRad(-1.068, 2.615));
	list.push_back(constRad(-0.800, 1.824));
	list.push_back(constRad(-1.316, 1.593));
	list.push_back(constRad(0.101, 1.316));
	list.push_back(constRad(-0.332, 1.026));
	list.push_back(constRad(-0.022, 2.069));
	list.push_back(constRad(-0.557, 2.399));
	list.push_back(constRad(-1.129, 0.626));
}

class ImageConverter
{
	ros::NodeHandle nh_;
	image_transport::ImageTransport it_;
	image_transport::Subscriber image_sub_;
	image_transport::Publisher thresh_image_pub_;
	image_transport::Publisher project_image_pub_;
	ros::Subscriber poseSub_;
	geometry_msgs::Pose currentPose_;
	geometry_msgs::Pose offsetPose_;
	std::mutex offsetPoseMutex_;
	tf::TransformListener poseListener_;
	std::vector<AMOS::IceFeature> iceFeatures_;
	int _imageDumpIdx;

	//AMOS::CameraModel camModel_;
	AMOS::RecursiveMapEstimator mapEstimator_;

	public:
	ImageConverter()
		: it_(nh_), mapEstimator_("ardrone_cameradata.xml"), _imageDumpIdx(0)
	{
		// Subscrive to input video feed and publish output video feed
		image_sub_ = it_.subscribe("/ardrone/bottom/image_raw", 1, 
				&ImageConverter::imageCb, this);
		thresh_image_pub_ = it_.advertise("/image_converter/output_threshold", 1);
		project_image_pub_ = it_.advertise("/image_converter/output_projected", 1);

		// Subscribe to pose messages
		poseSub_ = nh_.subscribe("/Robot_1/pose", 100, &ImageConverter::poseCb, this);
	
		// Create ice features
		createIceFeatureList(iceFeatures_);

		cv::namedWindow(OPENCV_WINDOW);
	}

	~ImageConverter()
	{
		cv::destroyWindow(OPENCV_WINDOW);
	}
	
	cv::Matx33f orientationToCamRot(const tf::Quaternion& bodyQuat)
	{
		tf::Quaternion bodyToCamQuat(0.707, -0.707, 0, 0);
		tf::Matrix3x3 rotMat(bodyToCamQuat*bodyQuat);
		return cv::Matx33f(rotMat[0][0], rotMat[0][1], rotMat[0][2], 
				   rotMat[1][0], rotMat[1][1], rotMat[1][2],
				   rotMat[2][0], rotMat[2][1], rotMat[2][2]);
	}

	/*
	cv::Matx33f orientationToCamRot(geometry_msgs::Quaternion quat)
	{
		tf::Quaternion bodyQuat(quat.x, quat.y, quat.z, quat.w);
		tf::Quaternion bodyToCamQuat(0.707, -0.707, 0, 0);
		tf::Matrix3x3 rotMat(bodyQuat*bodyToCamQuat);
		return cv::Matx33f(rotMat[0][0], rotMat[0][1], rotMat[0][2], 
				   rotMat[1][0], rotMat[1][1], rotMat[1][2],
				   rotMat[2][0], rotMat[2][1], rotMat[2][2]);
	}

	cv::Point3f posToPoint(geometry_msgs::Point pos)
	{
		return cv::Point3f(pos.x, pos.y, pos.z);
	}
	*/
	
	cv::Point3f posToPoint(tf::Vector3 pos)
	{
		tf::Quaternion bodyToCamQuat(0.707, -0.707, 0, 0);
		tf::Matrix3x3 rotMat(bodyToCamQuat);
		tf::Vector3 posCam = rotMat*pos;
		//auto posCam = pos;
		return cv::Point3f(posCam.x(), posCam.y(), posCam.z());
	}

	void poseCb(const geometry_msgs::PoseStamped::ConstPtr& poseStamped)
	{
		currentPose_ = poseStamped->pose;
	}

	void offsetPoseCb(geometry_msgs::PoseStamped pose)
	{
		std::this_thread::sleep_for(std::chrono::milliseconds(1000));
		std::lock_guard<std::mutex> lock(offsetPoseMutex_);
	}

	void imageCb(const sensor_msgs::ImageConstPtr& msg)
	{
		cv_bridge::CvImagePtr cv_thresh_ptr;
		cv_bridge::CvImagePtr cv_project_ptr;
		try
		{
			cv_thresh_ptr = cv_bridge::toCvCopy(msg, sensor_msgs::image_encodings::MONO8);
			cv_project_ptr = cv_bridge::toCvCopy(msg, sensor_msgs::image_encodings::BGR8);
		}
		catch (cv_bridge::Exception& e)
		{
			ROS_ERROR("cv_bridge exception: %s", e.what());
			return;
		}
		// Get image time stamp
		auto stamp = cv_thresh_ptr->header.stamp;
		ros::Duration delay(0.01);
		ros::Time imageTimestamp(stamp.sec, stamp.nsec);

		// Get transform at timestamp
		tf::StampedTransform bodyWorld;
		try {
			poseListener_.lookupTransform("/Robot_1/base_link", "/world", imageTimestamp-delay, bodyWorld);
		} catch(tf::TransformException &ex) {
			ROS_ERROR("%s", ex.what());
			return;
		}

		// Get pose
		cv::Matx33f dcmCamWorld = orientationToCamRot(bodyWorld.getRotation());
		cv::Point3f posWorld = posToPoint(bodyWorld.getOrigin());

		// Project the raw image onto the map plane for demonstration purposes
		cv::Mat mapImage;
		mapEstimator_.projectImageToMap(cv_project_ptr->image, mapImage, dcmCamWorld, posWorld);
		cv_project_ptr->image = mapImage;
		
		// Threshold the image
		cv::threshold(cv_thresh_ptr->image, cv_thresh_ptr->image, 200, 255, cv::THRESH_BINARY);
		cv::Mat segmented;
		cv::threshold(cv_thresh_ptr->image, segmented, 0, 1, cv::THRESH_BINARY);

		// Update map estimate (also projects image to map plane)
		cv::Mat segmented_32f;
		segmented.convertTo(segmented_32f, CV_32F);
		mapEstimator_.updateMap(segmented_32f, dcmCamWorld, posWorld);

		// Get probability map with features
		cv::Mat probMap;
		mapEstimator_.drawFeaturesOnMap(iceFeatures_, probMap);

		// Update GUI Window
		cv::Mat probMapScaled;
		cv::resize(probMap, probMapScaled, cv::Size(600,600));
		cv::imshow(OPENCV_WINDOW, probMapScaled);
		int retWaitKey = cv::waitKey(3);
		if(retWaitKey != -1) {
			cv::imwrite("dump"+std::to_string(_imageDumpIdx)+".png", 255*probMap);
			_imageDumpIdx++;
		}

		// Output modified video stream
		thresh_image_pub_.publish(cv_thresh_ptr->toImageMsg());
		project_image_pub_.publish(cv_project_ptr->toImageMsg());
	}
};

int main(int argc, char** argv)
{
	ros::init(argc, argv, "image_converter");
	ImageConverter ic;
	ros::spin();
	return 0;
}
