#include "gridmapping.hpp"

// ad-hoc error handling
#include <cstdlib>
#include <iostream>

namespace AMOS {

void 
fatal_err(const std::string& message) 
{
	std::cerr << message << std::endl;
	std::exit(1);
}

CameraModel::CameraModel(const std::string& paramFilename) 
{	
	cv::FileStorage fs(paramFilename, cv::FileStorage::READ);
	if(!fs.isOpened())
	{
		fatal_err("Unable to open camera parameters");
	}
	cv::Mat camMatrixTemp;
	fs["Camera_Matrix"] >> camMatrixTemp;
	camMatrix_ = camMatrixTemp;
	fs["Distortion_Coefficients"] >> distortionCoeff_;
}

void
CameraModel::undistortImage(const cv::Mat& input, cv::Mat& output)
{
	cv::undistort(input, output, camMatrix_, distortionCoeff_);
}

const cv::Matx33f 
GridMap::getMapWorldTransform()
{
	// Create homogeneous matrix representation of the map origin
	cv::Matx33f originHomogeneous(0,0,originWorldMap_.x,
				      0,0,originWorldMap_.y,
				      0,0,originWorldMap_.z);
	// Create coordinate transformation matrix from map to world
	cv::Matx33f rotScale = dcmWorldMap_*mapMatrix_;

	return rotScale + originHomogeneous;
}
		
void 
GridMap::addMeasurement(const cv::Mat measurement)
{
	map_ += measurement;
}

cv::Point3f 
GridMap::toMapFrame(cv::Point3f pointWorld) 
{
	cv::Mat pointWorldM(pointWorld, CV_32F);
	cv::Mat originM(originWorldMap_, CV_32F);
	cv::Mat dcmMapWorld(dcmWorldMap_.inv(), CV_32F);
	cv::Mat pointMapM = dcmMapWorld*pointWorldM - dcmMapWorld*originM;
	return cv::Point3f(pointMapM.at<float>(0), pointMapM.at<float>(1), pointMapM.at<float>(2));
}
		
void 
RecursiveMapEstimator::updateMap(cv::Mat& segmentedImage, const cv::Matx33f& dcmCamWorld, const cv::Point3f& originCamWorld)
{
	cv::Mat projectedMeasurement(gridMap_.getSize(), gridMap_.MAP_DTYPE, 0.0);
	cv::Mat inverseMeasurement;

	segmentedToProb(segmentedImage, inverseMeasurement);
	toLogOdds(inverseMeasurement, inverseMeasurement);
	projectImageToMap(inverseMeasurement, projectedMeasurement, dcmCamWorld, originCamWorld);
	cv::GaussianBlur(projectedMeasurement, projectedMeasurement, cv::Size(0,0), 1.0)
	gridMap_.addMeasurement(projectedMeasurement);
}

const cv::Matx33f 
RecursiveMapEstimator::getPerspectiveTransform(const cv::Matx33f dcmCamWorld, const cv::Point3f originCamWorld) 
{
	const cv::Matx33f camMat = camModel_.getCamMatrix();
	const cv::Matx33f mapWorldTransform = gridMap_.getMapWorldTransform();
	// Create homogeneous matrix representation of the camera origin
	cv::Matx33f camOriginHomogeneous(0,0,originCamWorld.x,
					 0,0,originCamWorld.y,
					 0,0,originCamWorld.z);
	
	// Create coordinate transformation from map to cam
	cv::Matx33f coordTransf = dcmCamWorld*mapWorldTransform;
	
	return (camMat*(coordTransf + camOriginHomogeneous)).inv();
}

void 
RecursiveMapEstimator::projectImageToMap(const cv::Mat& input, cv::Mat& output, const cv::Matx33f dcmCamWorld, const cv::Point3f originCamWorld)
{
	const cv::Matx33f tfM = getPerspectiveTransform(dcmCamWorld, originCamWorld);
	cv::Size mapSz = gridMap_.getSize();
	cv::warpPerspective(input, output, tfM, mapSz);
}

void
RecursiveMapEstimator::toLogOdds(cv::Mat& inputProb, cv::Mat& outputLogOdds)
{
	// Truncate values to [0,MAXPROB]
	const double MAXPROB = 0.99;
	cv::threshold(inputProb, outputLogOdds, MAXPROB, 0, cv::THRESH_TRUNC);
	cv::threshold(outputLogOdds, outputLogOdds, 0, 0, cv::THRESH_TOZERO);
	// Calculate log-odds
	cv::divide(outputLogOdds, 1-outputLogOdds, outputLogOdds);
	cv::log(outputLogOdds, outputLogOdds);
}

void 
RecursiveMapEstimator::toProbability(cv::Mat& inputLogOdds, cv::Mat& outputProbability)
{
	cv::exp(inputLogOdds, outputProbability);
	cv::divide(outputProbability, 1+outputProbability, outputProbability);
}


void
RecursiveMapEstimator::segmentedToProb(cv::Mat& segmentedImage, cv::Mat& detectionProb)
{
	const double Pd = 0.60;
	const double Pdf = 0.48;
	cv::threshold(segmentedImage, detectionProb, 0, Pd-Pdf, cv::THRESH_BINARY);
	detectionProb += Pdf;
}
		
void
RecursiveMapEstimator::getProbMap(cv::Mat& output)
{
	cv::Mat logoddsMap = gridMap_.getMap().clone();
	toProbability(logoddsMap, output);
}
		
void 
RecursiveMapEstimator::drawFeaturesOnMap(std::vector<IceFeature> features, cv::Mat& output) 
{
	getProbMap(output);
	cvtColor(output, output, CV_GRAY2RGB);
	double fontScale = 0.5; int fontFace = cv::FONT_HERSHEY_SIMPLEX;
	int index = 0;
	for(auto feature : features) {
		cv::Point3f loc = gridMap_.toMapFrame(feature.posWorld);
		cv::Point2i mapLoc = gridMap_.toMapCoordinates(cv::Point2f(loc.x, loc.y));
		cv::circle(output, mapLoc, 0, CV_RGB(255, 0, 0));
		cv::putText(output, std::to_string(index), mapLoc, fontFace, fontScale, cv::Scalar::all(255));
		index++;
	}
}

} // namespace AMOS
