#include "mavlink_interface.h"

namespace Snakelab {
// ardupilot MAVLINK copter modes
enum 
APM_copterModes 
{
	CP_MODE_STABILIZE = 0,                     // hold level position
	CP_MODE_ACRO = 1,                          // rate control
	CP_MODE_ALT_HOLD = 2,                      // AUTO control
	CP_MODE_AUTO = 3,                          // AUTO control
	CP_MODE_GUIDED = 4,                        // AUTO control
	CP_MODE_LOITER = 5,                        // Hold a single location
	CP_MODE_RTL = 6,                           // AUTO control
	CP_MODE_CIRCLE = 7,                        // AUTO control
	CP_MODE_POSITION = 8,                      // AUTO control
	CP_MODE_LAND = 9,                          // AUTO control
	CP_MODE_OF_LOITER = 10,                    // Hold a single location using optical flow sensor
	CP_MODE_DRIFT = 11,                        // DRIFT mode (Note: 12 is no longer used)
	CP_MODE_DUNE = 12,
	CP_MODE_SPORT = 13                         // earth frame rate control
};

MavlinkInterfaceNode::MavlinkInterfaceNode() 
: world_frame_ned("world_ned"), world_frame_ecef("world_ecef"), body_frame("body_aero"), world_frame_mocap("world")
{
	if(n.getParam("/mavlink_interface/url", url)) {
		ROS_INFO_STREAM("Loaded url parameter " << url);
	} else {
		ROS_INFO("Could not load url, shutting down");
		ros::shutdown();
	}
	
	boot_time = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::system_clock::now().time_since_epoch());
	mavInterfacePtr = mavconn::MAVConnInterface::open_url(url);
	mavInterfacePtr->message_received.connect(boost::bind(&MavlinkInterfaceNode::recv_message, this, _1, _2, _3));
	cmdPub = n.advertise<std_msgs::String>("tum_ardrone/com", 100);
}

void 
MavlinkInterfaceNode::run() 
{
	send_heartbeat();
	auto now = std::chrono::system_clock::now().time_since_epoch();
	auto time_elapsed = std::chrono::duration_cast<std::chrono::milliseconds>(now-boot_time).count();

	// Get and publish pose
	tf::StampedTransform ned_body_tf, ecef_body_tf;
	try {
		poseListener.lookupTransform(world_frame_ned, body_frame, ros::Time(0), ned_body_tf);
		poseListener.lookupTransform(world_frame_ecef, body_frame, ros::Time(0), ecef_body_tf);
		send_local_pose(ned_body_tf, static_cast<uint32_t>(time_elapsed));
		send_global_position(ecef_body_tf, static_cast<uint32_t>(time_elapsed));
		send_nav_controller_packet();
	} catch(tf::TransformException &ex) {
		ROS_ERROR("%s", ex.what());
		ros::Duration(1.0).sleep();
	}
}

void 
MavlinkInterfaceNode::recv_message(const mavlink_message_t *msg, uint8_t sysid, uint8_t compid) 
{
	switch(msg->msgid) {
		case MAVLINK_MSG_ID_MISSION_ITEM: 
		{
			mavlink_mission_item_t missionItem;
			mavlink_msg_mission_item_decode(msg, &missionItem);
			double lat = missionItem.x;
			double lon = missionItem.y;
			double hae = missionItem.z;
			double x_ecef, y_ecef, z_ecef;
			llh_to_ECEF(lat, lon, hae, x_ecef, y_ecef, z_ecef);
			tf::Stamped<tf::Point> pos_ecef(tf::Point(x_ecef, y_ecef, z_ecef), ros::Time(0), world_frame_ecef);
			tf::Stamped<tf::Point> pos_world;
			poseListener.transformPoint(world_frame_mocap, pos_ecef, pos_world);

			ROS_INFO("RECV MAVLINK MISSION ITEM, lat %f, lon %f, hae %f, w.x: %f, w.y %f, w.z %f", 
				missionItem.x, missionItem.y, missionItem.z,
				pos_world.getX(), pos_world.getY(), pos_world.getZ());

			std_msgs::String commandMsg;
			std::stringstream ss;
			ss << "c goto " << pos_world.getX() << " " << pos_world.getY() << " 1 0";
			commandMsg.data = ss.str();
			cmdPub.publish(commandMsg);

			send_waypoint_ack();

			break;
		}
		default:
		{
			ROS_INFO("RECV UNHANDLED MAVLINK MSG %d %d", msg->msgid, msg->seq);
			break;
		}
	}
}

void 
MavlinkInterfaceNode::send_heartbeat() 
{
	auto ip = mavInterfacePtr.get();
	mavlink_message_t msg;
	mavlink_msg_heartbeat_pack_chan(ip->get_system_id(), ip->get_component_id(), ip->get_channel(), &msg,
			MAV_TYPE_QUADROTOR,
			MAV_AUTOPILOT_ARDUPILOTMEGA,
			MAV_MODE_GUIDED_ARMED, // mode
			CP_MODE_GUIDED, // custom mode
			MAV_STATE_ACTIVE);
	ip->send_message(&msg);
}

void 
MavlinkInterfaceNode::send_nav_controller_packet() 
{
	auto ip = mavInterfacePtr.get();
	mavlink_message_t navMsg;
	mavlink_msg_nav_controller_output_pack(ip->get_system_id(), ip->get_component_id(), 
		&navMsg,
		0, // roll des 
		0, // pitch des
		0, // bearing des
		0, // target bearing
		0, // wp dist
		0, // alt error
		0, // aspd error
		0); // xtrack error
	ip->send_message(&navMsg);
}

void
MavlinkInterfaceNode::send_waypoint_ack()
{	
	auto ip = mavInterfacePtr.get();
	mavlink_message_t ackMsg;
	mavlink_msg_command_ack_pack(ip->get_system_id(), ip->get_component_id(),
		&ackMsg,
		MAV_CMD_NAV_WAYPOINT,
		MAV_RESULT_ACCEPTED);
	ip->send_message(&ackMsg);
}


void 
MavlinkInterfaceNode::send_global_position(const tf::StampedTransform& transform, uint32_t time_musec)
{
	auto ip = mavInterfacePtr.get();
	const tf::Vector3 trans = transform.getOrigin();
	double lat, lon, hae;
	mavlink_message_t position_msg;
	mavlink_message_t raw_msg;
	mavlink_message_t hud_msg;
	
	ECEF_to_llh(trans.x(), trans.y(), trans.z(), lat, lon, hae);
	mavlink_msg_global_position_int_pack(ip->get_system_id(), ip->get_component_id(),
		&position_msg,
		time_musec,
		static_cast<uint32_t>(lat*1E7),
		static_cast<uint32_t>(lon*1E7),
		static_cast<uint32_t>(hae*1E3), // should actually be AMSL according to MAVLINK
		static_cast<uint32_t>(hae*1E3), // should actually be relative alt
		0, // vx
		0, // vy
		0, // vz
		UINT16_MAX);
	mavlink_msg_gps_raw_int_pack(ip->get_system_id(), ip->get_component_id(),
		&raw_msg,
		time_musec,
		3, // GPS 3D fix
		static_cast<uint32_t>(lat*1E7),
		static_cast<uint32_t>(lon*1E7),
		static_cast<uint32_t>(hae*1E3), // should actually be AMSL according to MAVLINK
		UINT16_MAX, //eph
		UINT16_MAX, //epv
		UINT16_MAX, //vel
		UINT16_MAX, //cog
		255); //nsat
	mavlink_msg_vfr_hud_pack(ip->get_system_id(), ip->get_component_id(),
		&hud_msg,
		0.0, // airspeed
		0.0, // groundspeed
		0, // heading
		0, // throttle
		hae, // altitude
		0.0); // climb

	ip->send_message(&hud_msg);
	ip->send_message(&raw_msg);
	ip->send_message(&position_msg);
}

void 
MavlinkInterfaceNode::send_local_pose(const tf::StampedTransform& transform, uint32_t time_musec) 
{
	auto ip = mavInterfacePtr.get();
	const tf::Quaternion quat = transform.getRotation();
	const tf::Vector3 trans = transform.getOrigin();
	const tf::Vector3 q_imag = quat.getAxis();
	const tf::Matrix3x3 rotMat = tf::Matrix3x3(quat);
	auto q_real = quat.getW();
	mavlink_message_t orientation_quat_msg;
	mavlink_message_t orientation_euler_msg;
	mavlink_message_t position_ned_msg;
	mavlink_message_t vicon_msg;
	double roll, pitch, yaw;
	rotMat.getEulerZYX(yaw, pitch, roll); // this function is deprecated :(

	// Attitude quat
	mavlink_msg_attitude_quaternion_pack(ip->get_system_id(), ip->get_component_id(), 
		&orientation_quat_msg,
		time_musec,
		static_cast<float>(q_real),
		static_cast<float>(q_imag.x()),
		static_cast<float>(q_imag.y()),
		static_cast<float>(q_imag.z()),
		0.0, // roll speed
		0.0, // pitch speed
		0.0); // yaw speed
	// Attitude euler, ZYX convention	
	mavlink_msg_attitude_pack(ip->get_system_id(), ip->get_component_id(),
		&orientation_euler_msg,
		time_musec,
		roll,
		pitch,
		yaw,
		0.0, // roll speed
		0.0, // pitch speed
		0.0); // yaw speed
	// Local position
	mavlink_msg_local_position_ned_pack(ip->get_system_id(), ip->get_component_id(),
		&position_ned_msg,
		time_musec,
		static_cast<float>(trans.x()),
		static_cast<float>(trans.y()),
		static_cast<float>(trans.z()),
		0.0, // vx
		0.0, // vy
		0.0); // vz
	mavlink_msg_vicon_position_estimate_pack(ip->get_system_id(), ip->get_component_id(),
		&vicon_msg,
		time_musec,
		trans.x(),
		trans.y(),
		trans.z(),
		roll,
		pitch,
		yaw);
	ip->send_message(&orientation_quat_msg);
	ip->send_message(&orientation_euler_msg);
	ip->send_message(&position_ned_msg);
	ip->send_message(&vicon_msg);
}

void 
ECEF_to_llh(const double &x, const double &y, const double &z, double &lat_deg, double &lon_deg, double &hae)
{
	const double tol = 1E-15;
	const int max_it = 100;

	auto comp_l = [] (double x, double y) { return std::atan(y/x); };
	auto comp_mu0 = [] (double z, double p, double e) { return std::atan((z/p)/(1-std::pow(e,2))); };
	auto comp_N = [] (double re, double rp, double mu0) { return std::pow(re,2)/std::sqrt(std::pow(re*std::cos(mu0),2) + std::pow(rp*std::sin(mu0),2)); };
	auto comp_h = [] (double p, double mu0, double N) { return ((p/std::cos(mu0)) - N); };
	auto comp_mu = [] (double z, double p, double e, double N, double h) { return std::atan((z/p)/(1-std::pow(e,2)*(N/(N+h)))); };

	int it_num = 0;
	const double l = comp_l(x, y);
	const double e = std::sqrt(1 - std::pow(WGS84_RP/WGS84_RE,2));
	double p = std::sqrt(std::pow(x,2) + std::pow(y,2));
	double mu0 = comp_mu0(z, p, e);
	double mu = mu0;
	double N, h;
	do {
		it_num++;
		mu0 = mu;
		N = comp_N(WGS84_RE, WGS84_RP, mu0);
		h = comp_h(p, mu0, N);
		mu = comp_mu(z, p, e, N, h);
	} while(std::abs(mu - mu0) > tol && it_num < max_it);

	lat_deg = 180*mu/M_PI;
	lon_deg = 180*l/M_PI;
	hae = h;
}

void 
llh_to_ECEF(const double &lat_deg, const double &lon_deg, const double &hae, double &x, double &y, double &z)
{
	auto comp_f = [] (double a, double b) { return (a - b)/a; };
	auto comp_esq = [] (double f) { return 2*f - std::pow(f, 2); };
	auto comp_N = [] (double a, double esq, double lat_rad) { return a/std::sqrt(1 - esq*std::pow(std::sin(lat_rad), 2)); };

	double lat_rad = M_PI*lat_deg/180;
	double lon_rad = M_PI*lon_deg/180;
	double f = comp_f(WGS84_RE, WGS84_RP);
	double esq = comp_esq(f);
	double N = comp_N(WGS84_RE, esq, lat_rad);

	x = (N + hae)*std::cos(lat_rad)*std::cos(lon_rad);
	y = (N + hae)*std::cos(lat_rad)*std::sin(lon_rad);
	z = (N*(1 - esq) + hae)*std::sin(lat_rad);
}

} // namespace Snakelab
